"""
.. include:: ../README.md
.. include:: ../CHANGELOG.md
.. include:: ../pdocs/user_guide/1_concepts.md
.. include:: ../pdocs/user_guide/2_credentials.md
.. include:: ../pdocs/user_guide/3_postprocess.md
.. include:: ../pdocs/user_guide/4_pipelines.md
.. include:: ../pdocs/user_guide/5_data_management.md
.. include:: ../pdocs/developer_guide/1_main.md
"""
